/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 * 
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.edge;

import java.text.NumberFormat;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

public class DebugUtils {
	
	private static AtomicInteger polygon3dCounter = new AtomicInteger(1);
	private static AtomicInteger coordListCounter = new AtomicInteger(1);
	
	private DebugUtils() {
	}

	public static void printPolyLineSegment2dList(List<PolyLineSegment2d> rSegments) {
		Locale.setDefault(Locale.US);
		for (PolyLineSegment2d seg : rSegments) {
			Point2d start = seg.getStart().getPoint();
			Point2d end = seg.getEnd().getPoint();
			System.out.println(String.format(
					"segmentList.push_back(new PolyLineSegment2d(new Coordinate2d(%f, %f), new Coordinate2d(%f, %f)));",
					start.getX(), start.getY(), end.getX(), end.getY()));
		}
	}

	public static void printPolygon2d(Polygon2d mcpPolygon1) {
		Locale.setDefault(Locale.US);
		System.out.println("Coordinate2dList coords;");
		HalfEdge2d firstHalfEdge = mcpPolygon1.getFirstHalfEdge();
		Point2d start = firstHalfEdge.getStart().getPoint();
		System.out.println(String.format("coords.push_back(new Coordinate2d(%f, %f));", start.getX(), start.getY()));
		HalfEdge2d next = firstHalfEdge.getNext();
		while (next != null && next != firstHalfEdge) {
			start = next.getStart().getPoint();
			System.out.println(String.format("coords.push_back(new Coordinate2d(%f, %f));", start.getX(), start.getY()));
			next = next.getNext();
		}
		System.out.println("Polygon2d* mcpPolygon1 = new Polygon2dNs(coords);");
	}

	public static void printPolygon3d(EdgePolygon p) {
		Locale.setDefault(Locale.US);
		int counter = polygon3dCounter.getAndIncrement();
		System.out.println("Coordinate3dList coords" + counter + ";");
		HalfEdge firstHalfEdge = p.getFirstHalfEdge();
		Point3d start = firstHalfEdge.getStart().getPoint();
		System.out.println(String.format("coords%d.push_back(new Coordinate3d(%f, %f, %f));", counter, start.getX(), start.getY(), start.getZ()));
		HalfEdge next = firstHalfEdge.getNext();
		while (next != null && next != firstHalfEdge) {
			start = next.getStart().getPoint();
			System.out.println(String.format("coords%d.push_back(new Coordinate3d(%f, %f, %f));", counter, start.getX(), start.getY(), start.getZ()));
			next = next.getNext();
		}
		System.out.println("PolygonNs* p" + counter + " = new PolygonNs(coords" + counter+ ");");
	}

	public static void printPolygon3d(CDPolygonNs edgePoly1, CDPolygonNs edgePoly2) {
		Locale.setDefault(Locale.US);
		Map<Point3d, String> pointMap = new IdentityHashMap<>();
		int startCounter = 1;
		startCounter = printPolygon3d(edgePoly1, pointMap, startCounter);
		printPolygon3d(edgePoly2, pointMap, startCounter);
	}
	
	public static void printGeoknechtPolygon(CDPolygonNs... polys) {
		Locale.setDefault(Locale.US);
		NumberFormat nf = NumberFormat.getNumberInstance();
		nf.setMaximumFractionDigits(30);
		for (CDPolygonNs poly : polys) {
			List<Coordinate3d> coordinates = poly.getCoordinates();
			System.out.print("polygon(");
			for (Coordinate3d coord : coordinates) {
				Point3d point = coord.getPoint();
				writeGeoknechtPoint(nf, point);
			}
			writeGeoknechtPoint(nf, coordinates.get(0).getPoint());
			System.out.println(")");
			for (List<HalfEdge> innerRing : poly.getInnerHalfEdges()) {
				System.out.print("polygon(");
				for (HalfEdge he : innerRing) {
					Point3d point = he.getStart().getPoint();
					writeGeoknechtPoint(nf, point);
				}
				writeGeoknechtPoint(nf, innerRing.get(innerRing.size() - 1).getEnd().getPoint());
				System.out.println(")");
			}
		}
	}
	
	public static void printGeoknechtPolygon(EdgePolygon... polys) {
		Locale.setDefault(Locale.US);
		NumberFormat nf = NumberFormat.getNumberInstance();
		nf.setMaximumFractionDigits(30);
		for (EdgePolygon poly : polys) {
			List<Coordinate3d> coordinates = poly.getCoordinates();
			System.out.print("polygon(");
			for (Coordinate3d coord : coordinates) {
				Point3d point = coord.getPoint();
				writeGeoknechtPoint(nf, point);
			}
			writeGeoknechtPoint(nf, coordinates.get(0).getPoint());
			System.out.println(")");
		}
	}

	private static void writeGeoknechtPoint(NumberFormat nf, Point3d point) {
		System.out.print(nf.format(point.getX()));
		System.out.print("|");
		System.out.print(nf.format(point.getY()));
		System.out.print("|");
		System.out.print(nf.format(point.getZ()));
		System.out.print(" ");
	}

	private static int printPolygon3d(CDPolygonNs p, Map<Point3d, String> pointMap, int startCounter) {
		NumberFormat nf = NumberFormat.getNumberInstance();
		nf.setMaximumFractionDigits(30);
		int counter = polygon3dCounter.getAndIncrement();
		int listCounter = coordListCounter.getAndIncrement();
		int firstListCounter = listCounter;
		HalfEdge firstHalfEdge = p.getFirstHalfEdge();
		int[] counterArr = new int[] {startCounter};
		System.out.println("Coordinate3dList coords" + firstListCounter + ";");
		writeRing(pointMap, nf, listCounter, firstHalfEdge, counterArr);
		for (List<HalfEdge> innerRing : p.getInnerHalfEdges()) {
			listCounter = coordListCounter.getAndIncrement();
			System.out.println("Coordinate3dList coords" + listCounter + ";");
			writeRing(pointMap, nf, listCounter, innerRing.get(0), counterArr);
		}
		int lastListCounter = coordListCounter.get();
		System.out.println("std::list<Coordinate3dList> coordLists" + counter + ";");
		for (int i = firstListCounter; i < lastListCounter; i++) {
			System.out.println("coordLists" + counter + ".push_back(coords" + i +");");
		}
		System.out.println("CDPolygonNs* p" + counter + " = new CDPolygonNs(coordLists" + counter+ ");");
		return counterArr[0];
	}

	private static void writeRing(Map<Point3d, String> pointMap, NumberFormat nf, int counter, HalfEdge firstHalfEdge,
			int[] counterArr) {
		final Point3d start = firstHalfEdge.getStart().getPoint();
		String coordVarName = getCoordName(pointMap, nf, start, counterArr);
		System.out.println(String.format("coords%d.push_back(%s);", counter, coordVarName));
		HalfEdge next = firstHalfEdge.getNext();
		while (next != null && next != firstHalfEdge) {
			Point3d nextPoint = next.getStart().getPoint();
			coordVarName = getCoordName(pointMap, nf, nextPoint, counterArr);

			System.out.println(String.format("coords%d.push_back(%s);", counter, coordVarName));
			next = next.getNext();
		}
	}

	private static String getCoordName(Map<Point3d, String> pointMap, NumberFormat nf, final Point3d start,
			int[] counterArr) {
		String coordVarName = pointMap.computeIfAbsent(start, k -> {
			System.out.println(String.format("Coordinate3d* c%d = new Coordinate3d(%s, %s, %s);", counterArr[0], nf.format(start.getX()), nf.format(start.getY()), nf.format(start.getZ())));
			String name = "c" + counterArr[0];
			counterArr[0]++;
			return name;
		});
		return coordVarName;
	}

	public static void printPolygon3d(EdgePolygon polygon1, EdgePolygon polygon2) {
		Locale.setDefault(Locale.US);
		Map<Point3d, String> pointMap = new IdentityHashMap<>();
		int startCounter = 1;
		startCounter = printPolygon3d(polygon1, pointMap, startCounter);
		printPolygon3d(polygon2, pointMap, startCounter);
	}
	
	private static int printPolygon3d(EdgePolygon p, Map<Point3d, String> pointMap, int startCounter) {
		NumberFormat nf = NumberFormat.getNumberInstance();
		nf.setMaximumFractionDigits(30);
		int counter = polygon3dCounter.getAndIncrement();
		int listCounter = coordListCounter.getAndIncrement();
		int firstListCounter = listCounter;
		HalfEdge firstHalfEdge = p.getFirstHalfEdge();
		int[] counterArr = new int[] {startCounter};
		System.out.println("Coordinate3dList coords" + firstListCounter + ";");
		writeRing(pointMap, nf, listCounter, firstHalfEdge, counterArr);
		System.out.println("CDPolygonNs* p" + counter + " = new CDPolygonNs(coords" + firstListCounter + ");");
		return counterArr[0];
	}

	public static void printFPSelfIntersectionTestCaseCode(PolygonPolygonIntersection... intersectionResults){
		for (PolygonPolygonIntersection intersection : intersectionResults) {
			printFPSelfIntersectionTestCaseCode(intersection.getPolygon1(), intersection.getPolygon2());
		}
	}

	public static void printFPSelfIntersectionTestCaseCode(EdgePolygon polygon1, EdgePolygon polygon2) {
		Locale.setDefault(Locale.US);
		System.out.println("-----------------------------------------------------");
		System.out.println("");

		System.out.println("@Test");
		System.out.println("public void testFalsePositiveCASENAME() {");
		printTabbedLine("Geometry geom = new Geometry(GeometryType.SOLID, Lod.LOD1);");
		System.out.println("");

		printTestCasePolygonCode(polygon1, polygon2);


		printTabbedLine("CDPolygonNs edgePoly1 = CDPolygonNs.of(p1, new HashMap<>());");
		printTabbedLine("CDPolygonNs edgePoly2 = CDPolygonNs.of(p2, new HashMap<>());");
		printTabbedLine("List<PolygonPolygonIntersection> intersectPolygons;");
		printTabbedLine("intersectPolygons = IntersectPlanarPolygons.intersectPolygons(edgePoly1, edgePoly2, 0.000001, 0.001);");
		printTabbedLine("assertNotNull(intersectPolygons);");
		printTabbedLine("assertTrue(intersectPolygons.isEmpty());");
		System.out.println("");


		printTabbedLine("intersectPolygons = IntersectPlanarPolygons.intersectPolygons(edgePoly2, edgePoly1, 0.000001, 0.001);");
		printTabbedLine("assertNotNull(intersectPolygons);");
		printTabbedLine("assertTrue(intersectPolygons.isEmpty());");
		System.out.println("}");
	}

	private static void printTabbedLine(String line){
		Locale.setDefault(Locale.US);
		String format = "\t%s\n";
		System.out.format(format, line);
	}

	private static void printTestCasePolygonCode(EdgePolygon polygon1, EdgePolygon polygon2) {
		printTabbedLine("Polygon p1 = new ConcretePolygon();");
		printTabbedLine("geom.addPolygon(p1);");
		printTabbedLine("LinearRing ext = new LinearRing(LinearRingType.EXTERIOR);");
		printTabbedLine("p1.setExteriorRing(ext);");
		printTabbedLine("List<Vertex> vertices= new ArrayList<>();");
		printTestCasePolygonVertices(polygon1);
		System.out.println("");

		printTabbedLine("Polygon p2 = new ConcretePolygon();");
		printTabbedLine("geom.addPolygon(p2);");
		printTabbedLine("ext = new LinearRing(LinearRingType.EXTERIOR);");
		printTabbedLine("p2.setExteriorRing(ext);");
		printTabbedLine("vertices= new ArrayList<>();");
		printTestCasePolygonVertices(polygon2);
		System.out.println("");

	}

	private static void printTestCasePolygonVertices(EdgePolygon polygon) {
		List<Coordinate3d> coordList = polygon.getCoordinates();
		NumberFormat nf = NumberFormat.getNumberInstance();
		nf.setMaximumFractionDigits(30);

		String vertexFormat = "vertices.add(new Vertex(%s,%s,%s));";
        for (Coordinate3d coordinate3d : coordList) {
            Point3d point = coordinate3d.getPoint();
            printTabbedLine(String.format(vertexFormat,
                    nf.format(point.getX()), nf.format(point.getY()), nf.format(point.getZ())));
        }
		// close ring
		Point3d point = coordList.get(0).getPoint();
		printTabbedLine(String.format(vertexFormat,point.getX(),point.getY(),point.getZ()));
		printTabbedLine("ext.addAllVertices(vertices);");
	}


	public static int printCityDoctorPolygons(Polygon... polygons) {
		Locale.setDefault(Locale.US);
		NumberFormat nf = NumberFormat.getNumberInstance();
		nf.setMaximumFractionDigits(30);
		nf.setGroupingUsed(false);
		Map<Vertex, String> pointMap = new HashMap<>();
		AtomicInteger counter = new AtomicInteger(0);
		int listCounter = 0;
		for (Polygon p : polygons) {
			LinearRing ext = p.getExteriorRing();
			System.out.println("Coordinate3dList coords" + listCounter + ";");
			for (int i = 0; i < ext.getVertices().size() - 1; i++) {
				Vertex v = ext.getVertices().get(i);
				String coordName = pointMap.computeIfAbsent(v, key -> {
					int index = counter.getAndIncrement();
					System.out.println(String.format("Coordinate3d* c%d = new Coordinate3d(%s, %s, %s);", index, nf.format(v.getX()), nf.format(v.getY()), nf.format(v.getZ())));
					return "c" + index;
				});
				System.out.println(String.format("coords%d.push_back(%s);", listCounter, coordName));
			}
			System.out.println("CDPolygonNs* p" + listCounter + " = new CDPolygonNs(coords" + listCounter + ");");
			listCounter++;
		}
		return listCounter;
	}
	
	public static void printCityDoctorGeometry(Geometry geom) {
		int listCounter = printCityDoctorPolygons(geom.getPolygons().toArray(new Polygon[0]));
		System.out.println("CDPolygonNsList polygons;");
		for (int i = 0; i < listCounter; i++) {
			System.out.println("polygons.push_back(p" + i + ");");
		}
		System.out.println("CDMeshSurfaceNs* surface = new CDMeshSurfaceNs(polygons);");
	}

}
