package de.hft.stuttgart.citydoctor2.webservice.endpoints;

import java.io.BufferedInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.atomic.AtomicLong;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import javax.persistence.EntityManager;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;

import org.apache.commons.io.output.ByteArrayOutputStream;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.glassfish.jersey.media.multipart.FormDataContentDisposition;
import org.glassfish.jersey.media.multipart.FormDataParam;

import de.hft.stuttgart.citydoctor2.datastructure.AbstractBuilding;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurfaceType;
import de.hft.stuttgart.citydoctor2.datastructure.BoundingBox;
import de.hft.stuttgart.citydoctor2.datastructure.Building;
import de.hft.stuttgart.citydoctor2.datastructure.BuildingPart;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.Installation;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.math.Triangle3d;
import de.hft.stuttgart.citydoctor2.math.Vector3d;
import de.hft.stuttgart.citydoctor2.parser.CityGmlConsumer;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.tesselation.TesselatedPolygon;
import de.hft.stuttgart.citydoctor2.webservice.AuthenticationException;
import de.hft.stuttgart.citydoctor2.webservice.database.BuildingQueries;
import de.hft.stuttgart.citydoctor2.webservice.database.ModelQueries;
import de.hft.stuttgart.citydoctor2.webservice.model.BuildingReference;
import de.hft.stuttgart.citydoctor2.webservice.model.BuildingsReturnType;
import de.hft.stuttgart.citydoctor2.webservice.model.ModelReference;
import de.hft.stuttgart.citydoctor2.webservice.model.ParseStatus;
import de.hft.stuttgart.citydoctor2.webservice.model.ValidationReference;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.GmlFile;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.User;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.Validation;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.WebBuilding;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.WebCityDoctorModel;
import de.hft.stuttgart.citydoctor2.webservice.utils.AuthenticationUtils;
import de.hft.stuttgart.citydoctor2.webservice.utils.HibernateUtils;

@Path("models")
public class ModelsEndpoint {

	private static final String ERROR_WHILE_RETRIEVING_MODELS = "Error while retrieving models";

	private static final String ACCESS_VIOLATION = "User {} tried to access not owned model with id {}";

	private static final Logger logger = LogManager.getLogger(ModelsEndpoint.class);

	private static final AtomicLong FILE_COUNTER = new AtomicLong(0);
	private static final String DIGITS = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
	private static final int NAME_LENGTH = 40;
	private static final Random RANDOM = new Random();

	@GET
	@Produces(MediaType.APPLICATION_JSON)
	public Response getModels(@DefaultValue("0") @QueryParam("from") int from, @Context HttpServletRequest request) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			if (user == null) {
				return Response.status(Status.FORBIDDEN).build();
			}

			List<ModelReference> models = new ArrayList<>();
			for (WebCityDoctorModel webModel : user.getOwnedModels()) {
				ModelReference ref = new ModelReference(webModel.getId(), webModel.getFileName(), webModel.getStatus());
				models.add(ref);
				for (Validation valid : webModel.getValidations()) {
					ref.getValidations().add(ValidationReference.of(valid));
				}
			}
			return Response.ok(models).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} catch (Exception e) {
			logger.fatal(ERROR_WHILE_RETRIEVING_MODELS, e);
			return Response.serverError().build();
		} finally {
			HibernateUtils.closeManager(manager);
		}
	}

	@POST
	@Consumes(MediaType.MULTIPART_FORM_DATA)
	@Produces(MediaType.APPLICATION_JSON)
	public Response addModel(@FormDataParam("file") InputStream fileStream,
			@FormDataParam("file") FormDataContentDisposition fileDetails, @Context HttpServletRequest request,
			@Context ServletContext context) {
		if (fileStream == null || fileDetails == null) {
			return Response.status(Status.BAD_REQUEST).entity("No file provided").build();
		} else if (fileDetails.getFileName().length() > 255) {
			return Response.status(Status.BAD_REQUEST).entity("File name is too long (> 255 characters").build();
		}

		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);

			final File tempDir = (File) context.getAttribute("javax.servlet.context.tempdir");
			final String tempFileName = String.valueOf(FILE_COUNTER.getAndIncrement());
			File tempFile = new File(tempDir.getAbsolutePath() + File.separator + tempFileName);
			Files.copy(fileStream, tempFile.toPath(), StandardCopyOption.REPLACE_EXISTING);
			String fileName = fileDetails.getFileName();

			WebCityDoctorModel model = new WebCityDoctorModel();
			model.setFileName(fileName);
			model.setOwner(user);
			model.setStatus(ParseStatus.PARSING);
			ModelQueries.storeModel(model, manager);

			handleNewModel(model, tempFile, tempDir, context);

			ModelReference ref = new ModelReference(model.getId(), model.getFileName(), model.getStatus());
			return Response.ok(ref).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} catch (Exception e) {
			logger.fatal("Error while handling uploaded file.", e);
			return Response.serverError().build();
		} finally {
			HibernateUtils.closeManager(manager);
		}
	}

	@GET
	@Path("{modelId}/error_buildings")
	@Produces(MediaType.APPLICATION_JSON)
	public Response getErrorBuildings(@PathParam("modelId") int modelId,
			@QueryParam("from") @DefaultValue("0") int from, @QueryParam("valid") Integer validationId,
			@Context HttpServletRequest request) {
		if (validationId == null) {
			return Response.status(Status.BAD_REQUEST).build();
		}
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			WebCityDoctorModel model = manager.find(WebCityDoctorModel.class, modelId);
			if (model.getOwner().getId() != user.getId()) {
				logger.warn(ACCESS_VIOLATION, user.getId(), model.getId());
				return Response.status(Status.FORBIDDEN).build();
			}
			List<BuildingReference> buildings;
			long maxCount = BuildingQueries.getErrorBuildingCountForModel(modelId, validationId, manager);
			buildings = BuildingQueries.getErrorBuildingsWithErrorIndicator(from, modelId, validationId, manager);
			BuildingsReturnType returnValue = new BuildingsReturnType();
			returnValue.setFrom(from);
			returnValue.setTo(Math.min(from + 25l, maxCount));
			returnValue.setBuildings(buildings);
			return Response.ok(returnValue).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} catch (Exception e) {
			logger.fatal(ERROR_WHILE_RETRIEVING_MODELS, e);
			return Response.serverError().build();
		} finally {
			HibernateUtils.closeManager(manager);
		}
	}

	@GET
	@Path("{modelId}/buildings")
	@Produces(MediaType.APPLICATION_JSON)
	public Response getBuildings(@PathParam("modelId") int modelId, @QueryParam("from") @DefaultValue("0") int from,
			@QueryParam("valid") Integer validationId, @Context HttpServletRequest request) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			WebCityDoctorModel model = manager.find(WebCityDoctorModel.class, modelId);
			if (model.getOwner().getId() != user.getId()) {
				logger.warn(ACCESS_VIOLATION, user.getId(), model.getId());
				return Response.status(Status.FORBIDDEN).build();
			}
			List<BuildingReference> buildings;
			long maxCount = BuildingQueries.getBuildingCountForModel(modelId, manager);
			if (validationId == null) {
				buildings = BuildingQueries.getBuildingsForModel(from, modelId, manager);
			} else {
				buildings = BuildingQueries.getBuildingsWithErrorIndicator(from, modelId, validationId, manager);
			}
			BuildingsReturnType returnValue = new BuildingsReturnType();
			returnValue.setFrom(from);
			returnValue.setTo(Math.min(from + 25l, maxCount));
			returnValue.setBuildings(buildings);
			return Response.ok(returnValue).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} catch (Exception e) {
			logger.fatal(ERROR_WHILE_RETRIEVING_MODELS, e);
			return Response.serverError().build();
		} finally {
			HibernateUtils.closeManager(manager);
		}
	}

	@DELETE
	@Path("{modelId}")
	public Response deleteModel(@PathParam("modelId") int modelId, @Context HttpServletRequest request) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			WebCityDoctorModel model = manager.find(WebCityDoctorModel.class, modelId);
			if (model.getOwner().getId() != user.getId()) {
				logger.warn(ACCESS_VIOLATION, user.getId(), model.getId());
				return Response.status(Status.FORBIDDEN).build();
			}
			if (model.getZipFile() != null && model.getZipFile().getFilePath() != null) {
				File zip = new File(model.getZipFile().getFilePath());
				Files.delete(zip.toPath());
			}
			manager.getTransaction().begin();
			manager.remove(model);
			manager.getTransaction().commit();
			return Response.ok().build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} catch (Exception e) {
			logger.fatal("Error while deleting model {}", modelId, e);
			return Response.serverError().build();
		} finally {
			HibernateUtils.closeManager(manager);
		}
	}

	private void handleNewModel(WebCityDoctorModel oldModel, File gmlFile, File tempDir, ServletContext context) {
		Thread t = new Thread(() -> {
			EntityManager manager = HibernateUtils.createManager();
			WebCityDoctorModel model = manager.merge(oldModel);
			File parseFile = storeModelAsZipFile(gmlFile, tempDir, context, manager, model);
			try {
				parseDataModel(parseFile, model, manager);
				ModelQueries.updateStatus(model, ParseStatus.OK, manager);
			} catch (UncheckedIOException | CityGmlParseException e) {
				logger.catching(e);
				ModelQueries.updateStatus(model, ParseStatus.ERROR, manager);
			} finally {
				manager.close();
				if (parseFile != null) {
					try {
						Files.delete(parseFile.toPath());
					} catch (IOException e1) {
						logger.catching(e1);
					}
				}
			}
		});
		t.setUncaughtExceptionHandler((thread, e) -> {
			logger.catching(e);
			EntityManager manager = HibernateUtils.createManager();
			WebCityDoctorModel model = manager.merge(oldModel);
			ModelQueries.updateStatus(model, ParseStatus.ERROR, manager);
			manager.close();
		});
		t.start();
	}

	/**
	 * Stores the gml file as a zip into the configured folder. Returns the unpacked
	 * file in a temporary directory.
	 * 
	 * @param gmlFile the incoming file
	 * @param tempDir the temporary folder for the unzipped files
	 * @param context servlet context for retrieving where to store the zip files
	 * @param manager the database access manager
	 * @param model   the associated data model for the file
	 * @return the unzipped file in a temporary directory
	 */
	private File storeModelAsZipFile(File gmlFile, File tempDir, ServletContext context, EntityManager manager,
			WebCityDoctorModel model) {
		boolean isZipFile = isZipFile(model.getFileName());
		File parseFile = gmlFile;
		if (!parseFile.canWrite()) {
			logger.warn("Cannot write the file, deletion will fail, before zip");
		}
		File zipFile = gmlFile;
		String storagePath = context.getInitParameter("fileStorage");
		File fileStorageDir = new File(storagePath);
		try {
			if (!isZipFile) {
				zipFile = createZipFile(gmlFile, fileStorageDir);
				if (!parseFile.canWrite()) {
					logger.warn("Cannot write the file, deletion will fail, after zip");
				}
			} else {
				File copy = getStorageFile(fileStorageDir);
				Files.move(zipFile.toPath(), copy.toPath());
				zipFile = copy;
				parseFile = unzipFile(zipFile, tempDir);
			}
		} catch (IOException e) {
			logger.catching(e);
			ModelQueries.updateStatus(model, ParseStatus.ERROR, manager);
		}
		GmlFile fileModel = new GmlFile();
		fileModel.setFilePath(zipFile.getAbsolutePath());
		ModelQueries.updateModelWithBlob(model, fileModel, manager);
		return parseFile;
	}

	private File unzipFile(File zipFile, File tempDir) throws IOException {
		try (ZipInputStream zis = new ZipInputStream(new BufferedInputStream(new FileInputStream(zipFile)))) {
			ZipEntry entry = zis.getNextEntry();
			if (entry == null) {
				throw new IllegalStateException("Zip contains no files");
			}
			final String tempFileName = String.valueOf(FILE_COUNTER.getAndIncrement());
			File tempFile = new File(tempDir.getAbsolutePath() + File.separator + tempFileName);
			Files.copy(zis, tempFile.toPath(), StandardCopyOption.REPLACE_EXISTING);
			if (zis.getNextEntry() != null) {
				throw new IllegalStateException("Zip contains more than one file");
			}
			return tempFile;
		}
	}

	private void parseDataModel(File file, WebCityDoctorModel model, EntityManager manager)
			throws CityGmlParseException {
		ParserConfiguration config = new ParserConfiguration(8, false);
		CityGmlConsumer coConsumer = new CityGmlConsumer() {
			
			@Override
			public void accept(CityObject co) {
				if (co instanceof Building) {
					try {
						mapBuilding(model, co, manager);
					} catch (IOException e) {
						throw new UncheckedIOException(e);
					}
				} else {
					throw new UnsupportedOperationException();
				}
				logger.debug(co.getGmlId().getGmlString());
			}
		};

		CityGmlParser.streamCityGml(file.getAbsolutePath(), config, coConsumer, null);
	}

	private void mapBuilding(WebCityDoctorModel model, CityObject co, EntityManager manager) throws IOException {
		Building b = (Building) co;
		Lod highestLod = findHighestLod(b);

		WebBuilding buildingRef = new WebBuilding();
		buildingRef.setModel(model);
		buildingRef.setGmlId(b.getGmlId().getGmlString());

		if (highestLod == null) {
			buildingRef.setViewData(new byte[] {});
		} else {
			List<Polygon> polys = collectPolygonsForLod(b, highestLod);
			if (polys.isEmpty()) {
				throw new IllegalStateException("Found empty geometries for Lod: " + highestLod);
			}
			byte[] geometryData;
			BoundingBox bbox = BoundingBox.of(polys);
			geometryData = createGeometryData(polys, bbox.getCenter());
			double width = Math.max(bbox.getWidth(), bbox.getDepth());
			buildingRef.setModelWidth(width);
			buildingRef.setViewData(geometryData);
		}

		BuildingQueries.persistBuilding(buildingRef, manager);
	}

	private byte[] createGeometryData(List<Polygon> polys, Vector3d center) throws IOException {
		byte[] geometryData;
		try (ByteArrayOutputStream baos = new ByteArrayOutputStream();
				DataOutputStream dos = new DataOutputStream(baos)) {
			for (Polygon p : polys) {
				Vector3d normal = p.calculateNormalNormalized();
				float red = 1.0f;
				float green = 1.0f;
				float blue = 1.0f;
				BoundarySurface surface = p.getPartOfSurface();
				if (surface != null && surface.getType() == BoundarySurfaceType.ROOF) {
					green = 0.0f;
					blue = 0.0f;
				}
				TesselatedPolygon tesselatedPolygon = p.tesselate();
				for (Triangle3d t : tesselatedPolygon.getTriangles()) {
					writeTriangle(dos, normal, red, green, blue, t, center);
				}
			}
			dos.flush();
			geometryData = baos.toByteArray();
		}
		return geometryData;
	}

	private void writeTriangle(DataOutputStream dos, Vector3d normal, float red, float green, float blue, Triangle3d t,
			Vector3d center) throws IOException {
		float normalX = (float) normal.getX();
		float normalY = (float) normal.getY();
		float normalZ = (float) normal.getZ();
		dos.writeFloat((float) (t.getP1().getX() - center.getX()));
		dos.writeFloat((float) (t.getP1().getY() - center.getY()));
		dos.writeFloat((float) (t.getP1().getZ() - center.getZ()));
		dos.writeFloat(normalX);
		dos.writeFloat(normalY);
		dos.writeFloat(normalZ);
		dos.writeFloat(red);
		dos.writeFloat(green);
		dos.writeFloat(blue);
		dos.writeFloat((float) (t.getP2().getX() - center.getX()));
		dos.writeFloat((float) (t.getP2().getY() - center.getY()));
		dos.writeFloat((float) (t.getP2().getZ() - center.getZ()));
		dos.writeFloat(normalX);
		dos.writeFloat(normalY);
		dos.writeFloat(normalZ);
		dos.writeFloat(red);
		dos.writeFloat(green);
		dos.writeFloat(blue);
		dos.writeFloat((float) (t.getP3().getX() - center.getX()));
		dos.writeFloat((float) (t.getP3().getY() - center.getY()));
		dos.writeFloat((float) (t.getP3().getZ() - center.getZ()));
		dos.writeFloat(normalX);
		dos.writeFloat(normalY);
		dos.writeFloat(normalZ);
		dos.writeFloat(red);
		dos.writeFloat(green);
		dos.writeFloat(blue);
	}

	private List<Polygon> collectPolygonsForLod(Building b, Lod highestLod) {
		List<Polygon> polygons = new ArrayList<>();
		collectAbstractBuildingPolygons(b, polygons, highestLod);
		for (BuildingPart bp : b.getBuildingParts()) {
			collectAbstractBuildingPolygons(bp, polygons, highestLod);
		}
		return polygons;
	}

	private void collectAbstractBuildingPolygons(AbstractBuilding ab, List<Polygon> polygons, Lod highestLod) {
		collectCityObjectPolygons(ab, polygons, highestLod);
		for (Installation bi : ab.getBuildingInstallations()) {
			collectCityObjectPolygons(bi, polygons, highestLod);
			for (BoundarySurface bs : bi.getBoundarySurfaces()) {
				collectCityObjectPolygons(bs, polygons, highestLod);
			}
		}
		for (BoundarySurface bs : ab.getBoundarySurfaces()) {
			collectCityObjectPolygons(bs, polygons, highestLod);
		}
	}

	private void collectCityObjectPolygons(CityObject co, List<Polygon> polygons, Lod highestLod) {
		for (Geometry geom : co.getGeometries()) {
			if (geom.getLod() == highestLod) {
				for (Polygon p : geom.getPolygons()) {
					if (!p.isLink()) {
						polygons.add(p);
					}
				}
			}
		}
	}

	private Lod findHighestLod(Building b) {
		Lod lod = findHighestAbstractBuildingLod(b);
		for (BuildingPart bp : b.getBuildingParts()) {
			Lod bpLod = findHighestAbstractBuildingLod(bp);
			lod = getHigherLod(lod, bpLod);
		}
		return lod;
	}

	private Lod findHighestAbstractBuildingLod(AbstractBuilding ab) {
		Lod lod = findHighestCityObjectLod(ab);
		for (Installation bi : ab.getBuildingInstallations()) {
			Lod biLod = findHighestCityObjectLod(bi);
			lod = getHigherLod(lod, biLod);
			for (BoundarySurface bs : bi.getBoundarySurfaces()) {
				Lod bsLod = findHighestCityObjectLod(bs);
				lod = getHigherLod(lod, bsLod);
			}
		}
		for (BoundarySurface bs : ab.getBoundarySurfaces()) {
			Lod bsLod = findHighestCityObjectLod(bs);
			lod = getHigherLod(lod, bsLod);
		}
		return lod;
	}

	private Lod getHigherLod(Lod lod, Lod bsLod) {
		if (lod == null) {
			lod = bsLod;
		} else {
			if (!lod.isHigher(bsLod)) {
				lod = bsLod;
			}
		}
		return lod;
	}

	private Lod findHighestCityObjectLod(CityObject co) {
		Lod highestLod = null;
		for (Geometry geom : co.getGeometries()) {
			if (highestLod == null) {
				highestLod = geom.getLod();
			} else {
				if (geom.getLod().isHigher(highestLod)) {
					highestLod = geom.getLod();
				}
			}
		}
		return highestLod;
	}

	private File createZipFile(File gmlFile, File storageDir) throws IOException {
		if (!gmlFile.exists() || gmlFile.isDirectory()) {
			throw new IllegalStateException("File does not exist or is a directory");
		}

		File zipFile = getStorageFile(storageDir);

		try (ZipOutputStream zipOut = new ZipOutputStream(new FileOutputStream(zipFile))) {
			ZipEntry entry = new ZipEntry(gmlFile.getName());
			zipOut.putNextEntry(entry);
			java.nio.file.Path path = gmlFile.toPath();
			Files.copy(path, zipOut);
		}
		return zipFile;
	}

	private boolean isZipFile(String fileName) {
		int lastIndex = fileName.lastIndexOf('.');
		if (lastIndex != -1) {
			return fileName.substring(lastIndex).equalsIgnoreCase("zip");
		}
		return false;
	}

	private File getStorageFile(File dir) {
		String name = generateRandomName();
		File f = new File(dir.getAbsolutePath() + File.separator + name);
		while (f.exists()) {
			name = generateRandomName();
			f = new File(dir.getAbsolutePath() + File.separator + name);
		}
		return f;
	}

	private String generateRandomName() {
		StringBuilder sb = new StringBuilder(NAME_LENGTH);
		for (int i = 0; i < NAME_LENGTH; i++) {
			sb.append(DIGITS.charAt(RANDOM.nextInt(DIGITS.length())));
		}
		return sb.toString();
	}

	@GET
	@Path("{modelId}/status")
	@Produces(MediaType.APPLICATION_JSON)
	public Response getStatus(@PathParam("modelId") int modelId, @Context HttpServletRequest request) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			if (user == null) {
				return Response.status(Status.FORBIDDEN).build();
			}
			WebCityDoctorModel model = manager.find(WebCityDoctorModel.class, modelId);
			if (model.getOwner().getId() != user.getId()) {
				return Response.status(Status.FORBIDDEN).build();
			}
			ModelReference ref = new ModelReference(model.getId(), model.getFileName(), model.getStatus());
			return Response.ok(ref).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} finally {
			HibernateUtils.closeManager(manager);
		}
	}

	@SuppressWarnings("unused")
	private String determineModelName(WebCityDoctorModel webModel) {
		String modelName = webModel.getFileName();
		int fileEndingIndex = modelName.lastIndexOf('.');
		if (fileEndingIndex != -1) {
			modelName = modelName.substring(0, fileEndingIndex);
		}
		return modelName;
	}
}
